part of fling_pickle;

/// A class that can be turned into a [Pickle].
abstract class Pickleable {
  Pickle asPickle();
}

/// A Memento-like abstraction of an object's state.
class Pickle {
  /// Constructs an empty [Pickle].
  Pickle() : _values = {};

  /// Reads a `String` from this [Pickle].
  ///
  /// This will throw [ArgumentError] if the key is empty or if the associated
  /// value is of a different type than expected.
  String readString(final String key) => _read(key);

  /// Reads a `bool` from this [Pickle].
  ///
  /// This will throw [ArgumentError] if the key is empty or if the associated
  /// value is of a different type than expected.
  bool readBool(final String key) => _read(key);

  /// Reads an `int` from this [Pickle].
  ///
  /// This will throw [ArgumentError] if the key is empty or if the associated
  /// value is of a different type than expected.
  int readInt(final String key) => _read(key);

  /// Reads a `double` from this [Pickle].
  ///
  /// This will throw [ArgumentError] if the key is empty or if the associated
  /// value is of a different type than expected.
  double readDouble(final String key) => _read(key);

  /// Reads a `Pickle` from this [Pickle].
  ///
  /// This will throw [ArgumentError] if the key is empty or if the associated
  /// value is of a different type than expected.
  Pickle readPickle(final String key) => Pickle._(_read(key));

  /// Reads a `Pickleable` from this [Pickle].
  ///
  /// This will throw [ArgumentError] if the key is empty or if the associated
  /// value is of a different type than expected.
  T readPickleable<T>(
          final String key, final T Function(Pickle pickle) depickler) =>
      depickler(readPickle(key));

  /// Reads a `List` of [Pickleable]s from this [Pickle].
  ///
  /// This will throw [ArgumentError] if the key is empty or if the associated
  /// value is of a different type than expected.
  List<T> readPickleables<T>(
          final String key, final T Function(Pickle pickle) depickler) =>
      _read<List<Map<String, dynamic>>>(key)
          .map<T>((final Map<String, dynamic> map) => depickler(Pickle._(map)))
          .toList();

  /// Reads a `List` of [Pickle]s from this [Pickle].
  ///
  /// This will throw [ArgumentError] if the key is empty or if the associated
  /// value is of a different type than expected.
  List<Pickle> readPickles(final String key) =>
      _read<List<Map<String, dynamic>>>(key).map(_from).toList();

  /// Places a `String` in this [Pickle].
  ///
  /// This method is chainable.
  Pickle withString(final String key, final String value) {
    _values[key] = value;
    return this;
  }

  /// Places a `bool` in this [Pickle].
  ///
  /// This method is chainable.
  Pickle withBool(final String key, final bool value) {
    _values[key] = value;
    return this;
  }

  /// Places an `int` in this [Pickle].
  ///
  /// This method is chainable.
  Pickle withInt(final String key, final int value) {
    _values[key] = value;
    return this;
  }

  /// Places a `double` in this [Pickle].
  ///
  /// This method is chainable.
  Pickle withDouble(final String key, final double value) {
    _values[key] = value;
    return this;
  }

  /// Places a `Pickle` in this [Pickle].
  ///
  /// This method is chainable.
  Pickle withPickle(final String key, final Pickle pickle) {
    _values[key] = pickle._values;
    return this;
  }

  /// Places a `Pickleable` in this [Pickle].
  ///
  /// This method is chainable.
  Pickle withPickleable(final String key, final Pickleable pickleable) =>
      withPickle(key, pickleable.asPickle());

  /// Places a `List` of [Pickle]s in this [Pickle].
  ///
  /// This method is chainable.
  Pickle withPickles(final String key, final List<Pickle> pickles) {
    _values[key] = pickles
        .map((final Pickle pickle) => pickle._values)
        .toList(growable: false);
    return this;
  }

  /// Places a `List` of [Pickleable]s in this [Pickle].
  ///
  /// This method is chainable.
  Pickle withPickleables(
      final String key, final Iterable<Pickleable> pickleables) {
    _values[key] = pickleables
        .map((final Pickleable pickleable) => pickleable.asPickle()._values)
        .toList(growable: false);
    return this;
  }

  /// Attempts to read an attribute from a specific key.
  ///
  /// This will throw [ArgumentError] if the key is empty or if the associated
  /// value is of a different type than expected.
  T _read<T>(final String key) {
    if (!_values.containsKey(key)) {
      throw ArgumentError('No such value: "$key"');
    }
    try {
      return _values[key];
    } on TypeError {
      throw ArgumentError('Incorrect value type: "$key"');
    }
  }

  static Pickle _from(final Map<String, dynamic> pickleValues) =>
      Pickle._(pickleValues);

  /// The values stored in this [Pickle].
  final Map<String, dynamic> _values;

  /// Constructs a [Pickle] from existing data.
  Pickle._(final Map<String, dynamic> pickleValues) : _values = pickleValues;
}
