import 'package:fling_pickle/fling_pickle.dart';
import 'package:test/test.dart';

class Tester implements Pickleable {
  final String name;

  Tester({final this.name = 'Fred'});

  static Tester fromPickle(final Pickle pickle) =>
      Tester(name: pickle.readString('name'));

  @override
  Pickle asPickle() {
    return Pickle().withString('name', name);
  }

  @override
  bool operator ==(final dynamic other) =>
      other is Tester && other.name == name;
}

void main() {
  group('withBool', () {
    test('assigned and read', () {
      // given
      final pickle = Pickle();
      pickle.withBool('key', true);

      // when
      final result = pickle.readBool('key');

      // then
      expect(result, true);
    });
    test('without assignment', () {
      // given
      final pickle = Pickle();

      // then
      expect(() => pickle.readBool('key'), throwsArgumentError);
    });
    test('assigned as different type', () {
      // given
      final pickle = Pickle();
      pickle.withString('key', 'true');

      // then
      expect(() => pickle.readBool('key'), throwsArgumentError);
    });
  });

  group('withDouble', () {
    test('assigned and read', () {
      // given
      final pickle = Pickle();
      pickle.withDouble('key', 2.4);

      // when
      final result = pickle.readDouble('key');

      // then
      expect(result, 2.4);
    });
    test('without assignment', () {
      // given
      final pickle = Pickle();

      // then
      expect(() => pickle.readDouble('key'), throwsArgumentError);
    });
    test('assigned as different type', () {
      // given
      final pickle = Pickle();
      pickle.withString('key', '2.4');

      // then
      expect(() => pickle.readDouble('key'), throwsArgumentError);
    });
  });

  group('withInt', () {
    test('assigned and read', () {
      // given
      final pickle = Pickle();
      pickle.withInt('key', 2);

      // when
      final result = pickle.readInt('key');

      // then
      expect(result, 2);
    });
    test('without assignment', () {
      // given
      final pickle = Pickle();

      // then
      expect(() => pickle.readInt('key'), throwsArgumentError);
    });
    test('assigned as different type', () {
      // given
      final pickle = Pickle();
      pickle.withString('key', '2');

      // then
      expect(() => pickle.readInt('key'), throwsArgumentError);
    });
  });

  group('withString', () {
    test('assigned and read', () {
      // given
      final pickle = Pickle();
      pickle.withString('key', 'SUCCESS');

      // when
      final result = pickle.readString('key');

      // then
      expect(result, 'SUCCESS');
    });
    test('without assignment', () {
      // given
      final pickle = Pickle();

      // then
      expect(() => pickle.readString('key'), throwsArgumentError);
    });
    test('assigned as different type', () {
      // given
      final pickle = Pickle();
      pickle.withInt('key', 2);

      // then
      expect(() => pickle.readString('key'), throwsArgumentError);
    });
  });

  group('withPickles', () {
    test('assigned and read', () {
      // given
      final pickle = Pickle();
      pickle.withPickles('key', [Pickle()]);

      // when
      final result = pickle.readPickles('key');

      // then
      expect(result.length, 1);
    });
    test('without assignment', () {
      // given
      final pickle = Pickle();

      // then
      expect(() => pickle.readPickles('key'), throwsArgumentError);
    });
    test('assigned as different type', () {
      // given
      final pickle = Pickle();
      pickle.withString('key', '[]');

      // then
      expect(() => pickle.readPickles('key'), throwsArgumentError);
    });
  });

  group('withPickle', () {
    test('assigned and read', () {
      // given
      final pickle = Pickle();
      pickle.withPickle('key', Pickle().withInt('lock', 2));

      // when
      final result = pickle.readPickle('key');

      // then
      expect(result.readInt('lock'), 2);
    });
    test('without assignment', () {
      // given
      final pickle = Pickle();

      // then
      expect(() => pickle.readPickle('key'), throwsArgumentError);
    });
    test('assigned as different type', () {
      // given
      final pickle = Pickle();
      pickle.withString('key', 'Pickle');

      // then
      expect(() => pickle.readPickle('key'), throwsArgumentError);
    });
  });

  group('withPickleable', () {
    test('assigned and read', () {
      // given
      final pickle = Pickle();
      pickle.withPickleable('key', Tester(name: 'kiwi'));

      // when
      final result = pickle.readPickleable('key', Tester.fromPickle);

      // then
      expect(result.name, 'kiwi');
    });
    test('without assignment', () {
      // given
      final pickle = Pickle();

      // then
      expect(() => pickle.readPickleable('key', Tester.fromPickle),
          throwsArgumentError);
    });
    test('assigned as different type', () {
      // given
      final pickle = Pickle();
      pickle.withString('key', 'Tester');

      // then
      expect(() => pickle.readPickleable('key', Tester.fromPickle),
          throwsArgumentError);
    });
  });

  group('withPickleables', () {
    test('assigned and read', () {
      // given
      final pickle = Pickle();
      pickle.withPickleables(
          'key', [Tester(name: 'Henry'), Tester(name: 'Henrietta')]);

      // when
      final result = pickle.readPickleables('key', Tester.fromPickle);

      // then
      expect(result, [Tester(name: 'Henry'), Tester(name: 'Henrietta')]);
    });
    test('without assignment', () {
      // given
      final pickle = Pickle();

      // then
      expect(() => pickle.readPickleables('key', Tester.fromPickle),
          throwsArgumentError);
    });
    test('assigned as different type', () {
      // given
      final pickle = Pickle();
      pickle.withString('key', '2');

      // then
      expect(() => pickle.readPickleables('key', Tester.fromPickle),
          throwsArgumentError);
    });
  });
}
